from pathlib import Path

DEFAULT_GETML_PROJECTS_PATH: Path = Path.home() / ".getML" / "projects"
DEFAULT_GETML_HOST: str = "localhost"
DEFAULT_GETML_PORT: int = 1709
DEFAULT_MLFLOW_TRACKING_URI: str = "http://localhost:5000"
